@@ -392,7 +392,7 @@ wl_connection_queue(struct wl_connection *connection,
 	return wl_buffer_put(&connection->out, data, count);
 }
 
-static int
+int
 wl_message_count_arrays(const struct wl_message *message)
 {
 	int i, arrays;
@@ -797,19 +797,6 @@ wl_connection_demarshal(struct wl_connection *connection,
 	return NULL;
 }
 
-int
-wl_interface_equal(const struct wl_interface *a, const struct wl_interface *b)
-{
-	/* In most cases the pointer equality test is sufficient.
-	 * However, in some cases, depending on how things are split
-	 * across shared objects, we can end up with multiple
-	 * instances of the interface metadata constants.  So if the
-	 * pointers match, the interfaces are equal, if they don't
-	 * match we have to compare the interface names. */
-
-	return a == b || strcmp(a->name, b->name) == 0;
-}
-
 int
 wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 {
@@ -838,7 +825,6 @@ wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 			} else if (object == NULL && id != 0) {
 				wl_log("unknown object (%u), message %s(%s)\n",
 				       id, message->name, message->signature);
-				object = NULL;
 				errno = EINVAL;
 				return -1;
 			}
