@@ -31,6 +31,7 @@
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdarg.h>
+#include <stdint.h>
 #include <string.h>
 #include <errno.h>
 #include <ctype.h>
@@ -49,6 +50,8 @@ extern int DTD_DATA_len;
 
 #include "wayland-util.h"
 
+#define PROGRAM_NAME "wayland-scanner"
+
 enum side {
 	CLIENT,
 	SERVER,
@@ -57,8 +60,8 @@ enum side {
 static int
 usage(int ret)
 {
-	fprintf(stderr, "usage: ./scanner [OPTION] [client-header|server-header|code]"
-		" [input_file output_file]\n");
+	fprintf(stderr, "usage: %s [OPTION] [client-header|server-header|code]"
+		" [input_file output_file]\n", PROGRAM_NAME);
 	fprintf(stderr, "\n");
 	fprintf(stderr, "Converts XML protocol descriptions supplied on "
 			"stdin or input file to client\n"
@@ -76,7 +79,7 @@ usage(int ret)
 static int
 scanner_version(int ret)
 {
-	fprintf(stderr, "wayland-scanner %s\n", WAYLAND_VERSION);
+	fprintf(stderr, "%s %s\n", PROGRAM_NAME, WAYLAND_VERSION);
 	exit(ret);
 }
 
@@ -210,6 +213,7 @@ struct enumeration {
 	struct wl_list link;
 	struct description *description;
 	bool bitfield;
+	int since;
 };
 
 struct entry {
@@ -217,6 +221,7 @@ struct entry {
 	char *uppercase_name;
 	char *value;
 	char *summary;
+	int since;
 	struct wl_list link;
 };
 
@@ -236,7 +241,7 @@ static void *
 fail_on_null(void *p)
 {
 	if (p == NULL) {
-		fprintf(stderr, "wayland-scanner: out of memory\n");
+		fprintf(stderr, "%s: out of memory\n", PROGRAM_NAME);
 		exit(EXIT_FAILURE);
 	}
 
@@ -491,6 +496,7 @@ create_enumeration(const char *name)
 	enumeration = xzalloc(sizeof *enumeration);
 	enumeration->name = xstrdup(name);
 	enumeration->uppercase_name = uppercase_dup(name);
+	enumeration->since = 1;
 
 	wl_list_init(&enumeration->entry_list);
 
@@ -574,8 +580,17 @@ free_interface(struct interface *interface)
 	free(interface);
 }
 
-/* convert string to unsigned integer,
- * in the case of error, return -1 */
+/* Convert string to unsigned integer
+ *
+ * Parses a non-negative base-10 number from the given string.  If the
+ * specified string is blank, contains non-numerical characters, is out
+ * of range, or results in a negative number, -1 is returned to indicate
+ * an error.
+ *
+ * Upon error, this routine does not modify or set errno.
+ *
+ * Returns -1 on error, or a non-negative integer on success
+ */
 static int
 strtouint(const char *str)
 {
@@ -597,6 +612,27 @@ strtouint(const char *str)
 	return (int)ret;
 }
 
+static int
+version_from_since(struct parse_context *ctx, const char *since)
+{
+	int version;
+
+	if (since != NULL) {
+		version = strtouint(since);
+		if (version == -1) {
+			fail(&ctx->loc, "invalid integer (%s)\n", since);
+		} else if (version > ctx->interface->version) {
+			fail(&ctx->loc, "since (%u) larger than version (%u)\n",
+			     version, ctx->interface->version);
+		}
+	} else {
+		version = 1;
+	}
+
+
+	return version;
+}
+
 static void
 start_element(void *data, const char *element_name, const char **atts)
 {
@@ -682,17 +718,7 @@ start_element(void *data, const char *element_name, const char **atts)
 		if (type != NULL && strcmp(type, "destructor") == 0)
 			message->destructor = 1;
 
-		if (since != NULL) {
-			version = strtouint(since);
-			if (version == -1) {
-				fail(&ctx->loc, "invalid integer (%s)\n", since);
-			} else if (version > ctx->interface->version) {
-				fail(&ctx->loc, "since (%u) larger than version (%u)\n",
-				     version, ctx->interface->version);
-			}
-		} else {
-			version = 1;
-		}
+		version = version_from_since(ctx, since);
 
 		if (version < ctx->interface->since)
 			warn(&ctx->loc, "since version not increasing\n");
@@ -774,6 +800,12 @@ start_element(void *data, const char *element_name, const char **atts)
 			fail(&ctx->loc, "no entry name given");
 
 		entry = create_entry(name, value);
+		version = version_from_since(ctx, since);
+
+		if (version < ctx->enumeration->since)
+			warn(&ctx->loc, "since version not increasing\n");
+		ctx->enumeration->since = version;
+		entry->since = version;
 
 		if (summary)
 			entry->summary = xstrdup(summary);
@@ -808,7 +840,7 @@ find_enumeration(struct protocol *protocol,
 	struct interface *i;
 	struct enumeration *e;
 	char *enum_name;
-	uint idx = 0, j;
+	uint32_t idx = 0, j;
 
 	for (j = 0; j + 1 < strlen(enum_attribute); j++) {
 		if (enum_attribute[j] == '.') {
@@ -962,7 +994,7 @@ emit_opcodes(struct wl_list *message_list, struct interface *interface)
 
 	opcode = 0;
 	wl_list_for_each(m, message_list, link)
-		printf("#define %s_%s\t%d\n",
+		printf("#define %s_%s %d\n",
 		       interface->uppercase_name, m->uppercase_name, opcode++);
 
 	printf("\n");
@@ -975,7 +1007,7 @@ emit_opcode_versions(struct wl_list *message_list, struct interface *interface)
 
 	wl_list_for_each(m, message_list, link) {
 		printf("/**\n * @ingroup iface_%s\n */\n", interface->name);
-		printf("#define %s_%s_SINCE_VERSION\t%d\n",
+		printf("#define %s_%s_SINCE_VERSION %d\n",
 		       interface->uppercase_name, m->uppercase_name, m->since);
 	}
 
@@ -1255,16 +1287,33 @@ emit_enumerations(struct interface *interface)
 		}
 		printf("enum %s_%s {\n", interface->name, e->name);
 		wl_list_for_each(entry, &e->entry_list, link) {
-			if (entry->summary)
-				printf("\t/**\n"
-				       "\t * %s\n"
-				       "\t */\n", entry->summary);
+			if (entry->summary || entry->since > 1) {
+				printf("\t/**\n");
+				if (entry->summary)
+					printf("\t * %s\n", entry->summary);
+				if (entry->since > 1)
+					printf("\t * @since %d\n", entry->since);
+				printf("\t */\n");
+			}
 			printf("\t%s_%s_%s = %s,\n",
 			       interface->uppercase_name,
 			       e->uppercase_name,
 			       entry->uppercase_name, entry->value);
 		}
 		printf("};\n");
+
+		wl_list_for_each(entry, &e->entry_list, link) {
+			if (entry->since == 1)
+                            continue;
+
+                        printf("/**\n * @ingroup iface_%s\n */\n", interface->name);
+                        printf("#define %s_%s_%s_SINCE_VERSION %d\n",
+                               interface->uppercase_name,
+                               e->uppercase_name, entry->uppercase_name,
+                               entry->since);
+
+		}
+
 		printf("#endif /* %s_%s_ENUM */\n\n",
 		       interface->uppercase_name, e->uppercase_name);
 	}
@@ -1350,7 +1399,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface, enum sid
 
 	if (side == CLIENT) {
 	    printf("/**\n"
-		   " * @ingroup %s_iface\n"
+		   " * @ingroup iface_%s\n"
 		   " */\n", interface->name);
 	    printf("static inline int\n"
 		   "%s_add_listener(struct %s *%s,\n"
@@ -1467,7 +1516,7 @@ emit_header(struct protocol *protocol, enum side side)
 	const char *s = (side == SERVER) ? "SERVER" : "CLIENT";
 	char **p, *prev;
 
-	printf("/* Generated by wayland-scanner %s */\n\n", WAYLAND_VERSION);
+	printf("/* Generated by %s %s */\n\n", PROGRAM_NAME, WAYLAND_VERSION);
 
 	printf("#ifndef %s_%s_PROTOCOL_H\n"
 	       "#define %s_%s_PROTOCOL_H\n"
@@ -1542,10 +1591,12 @@ emit_header(struct protocol *protocol, enum side side)
 			emit_structs(&i->request_list, i, side);
 			emit_opcodes(&i->event_list, i);
 			emit_opcode_versions(&i->event_list, i);
+			emit_opcode_versions(&i->request_list, i);
 			emit_event_wrappers(&i->event_list, i);
 		} else {
 			emit_structs(&i->event_list, i, side);
 			emit_opcodes(&i->request_list, i);
+			emit_opcode_versions(&i->event_list, i);
 			emit_opcode_versions(&i->request_list, i);
 			emit_stubs(&i->request_list, i);
 		}
@@ -1670,7 +1721,7 @@ emit_code(struct protocol *protocol)
 	struct wl_array types;
 	char **p, *prev;
 
-	printf("/* Generated by wayland-scanner %s */\n\n", WAYLAND_VERSION);
+	printf("/* Generated by %s %s */\n\n", PROGRAM_NAME, WAYLAND_VERSION);
 
 	if (protocol->copyright)
 		format_text_to_comment(protocol->copyright, true);
