@@ -30,10 +30,15 @@
 
 #include <stdarg.h>
 #include <stdlib.h>
+#include <stdint.h>
 
 #define WL_HIDE_DEPRECATED 1
 
 #include "wayland-util.h"
+#include "wayland-server-core.h"
+
+/* Invalid memory address */
+#define WL_ARRAY_POISON_PTR (void *) 4
 
 #define ARRAY_LENGTH(a) (sizeof (a) / sizeof (a)[0])
 
@@ -55,6 +60,10 @@ struct wl_object {
 extern struct wl_object global_zombie_object;
 #define WL_ZOMBIE_OBJECT ((void*)&global_zombie_object)
 
+int
+wl_interface_equal(const struct wl_interface *iface1,
+		   const struct wl_interface *iface2);
+
 /* Flags for wl_map_insert_new and wl_map_insert_at.  Flags can be queried with
  * wl_map_lookup_flags.  The current implementation has room for 1 bit worth of
  * flags.  If more flags are ever added, the implementation of wl_map will have
@@ -70,7 +79,8 @@ struct wl_map {
 	uint32_t free_list;
 };
 
-typedef void (*wl_iterator_func_t)(void *element, void *data);
+typedef enum wl_iterator_result (*wl_iterator_func_t)(void *element,
+						      void *data);
 
 void
 wl_map_init(struct wl_map *map, uint32_t side);
@@ -99,14 +109,6 @@ wl_map_lookup_flags(struct wl_map *map, uint32_t i);
 void
 wl_map_for_each(struct wl_map *map, wl_iterator_func_t func, void *data);
 
-struct wl_connection;
-struct wl_closure;
-struct wl_proxy;
-
-int
-wl_interface_equal(const struct wl_interface *iface1,
-		   const struct wl_interface *iface2);
-
 struct wl_connection *
 wl_connection_create(int fd);
 
@@ -161,6 +163,9 @@ get_next_argument(const char *signature, struct argument_details *details);
 int
 arg_count_for_signature(const char *signature);
 
+int
+wl_message_count_arrays(const struct wl_message *message);
+
 int
 wl_message_get_since(const struct wl_message *message);
 
@@ -229,4 +234,21 @@ zalloc(size_t s)
 	return calloc(1, s);
 }
 
+struct wl_priv_signal {
+	struct wl_list listener_list;
+	struct wl_list emit_list;
+};
+
+void
+wl_priv_signal_init(struct wl_priv_signal *signal);
+
+void
+wl_priv_signal_add(struct wl_priv_signal *signal, struct wl_listener *listener);
+
+struct wl_listener *
+wl_priv_signal_get(struct wl_priv_signal *signal, wl_notify_func_t notify);
+
+void
+wl_priv_signal_emit(struct wl_priv_signal *signal, void *data);
+
 #endif
